@@ -755,39 +755,6 @@ impl Rearrange {
                 )
             })?;
 
-        // check that known sizes of each axis divide the operand shape
-        let input_sizes: Vec<usize> = out
-            .spec
-            .input_ints
-            .iter()
-            .map(|x| {
-                x.iter()
-                    .map(|y| {
-                        let s = out.spec.int_sizes[*y as usize];
-                        if s.is_none() {
-                            1
-                        } else {
-                            s.unwrap()
-                        }
-                    })
-                    .product()
-            })
-            .collect();
-        let any_sizes_dont_divide = node.info().shape.iter().enumerate().any(|(i, l)| {
-            if input_sizes[i] == 0 {
-                *l == 0
-            } else {
-                *l % input_sizes[i] != 0
-            }
-        });
-
-        if input_sizes.len() != node.info().rank() || any_sizes_dont_divide {
-            return Err(ConstructError::RearrangeWrongInputShape {
-                spec: (*out.spec).clone(),
-                shape: node.info().shape.clone(),
-            }
-            .into());
-        }
         out.initial_init_info()
     }
 
