@@ -12,7 +12,9 @@ use rr_util::{
     cached_method,
     caching::FastUnboundedCache,
     eq_by_big_hash::EqByBigHash,
-    impl_eq_by_big_hash, pycall, python_error_exception,
+    impl_eq_by_big_hash,
+    name::Name,
+    pycall, python_error_exception,
     rearrange_spec::{ExpandToSpecOrShape, OpShape, OpSize, RearrangeSpec},
     sv,
     symbolic_size::SymbolicSizeProduct,
@@ -189,7 +191,7 @@ pub fn expand_node(
                         x.ndim(),
                     )
                     .unwrap(),
-                    x.name().map(|x| format!("{} rep_for_batch", x)),
+                    x.info().name.map(|x| format!("{} rep_for_batch", x).into()),
                 ))
             })
             .collect()
@@ -224,7 +226,7 @@ pub fn expand_node(
                 .to_smallvec();
             let c = inputs[0].clone();
             Ok(if s2 != *child_shape {
-                SetSymbolicShape::nrc(c, s2, s.name_cloned())
+                SetSymbolicShape::nrc(c, s2, s.info().name)
             } else {
                 c
             })
@@ -267,7 +269,7 @@ pub fn expand_node(
             };
 
             let new_spec = expanded_spec.add_batch_dims(batch_ranks[0]);
-            Rearrange::try_new(new_input, new_spec, circuit.name_cloned()).map(|x| x.rc())
+            Rearrange::try_new(new_input, new_spec, circuit.info().name).map(|x| x.rc())
         }
         // I think we just don't do any symbolic shape munging in Index/Scatter?
         Circuit::Index(node) => {
@@ -291,7 +293,7 @@ pub fn expand_node(
                         .chain(node.index.0.iter().cloned())
                         .collect(),
                 ),
-                node.name_cloned(),
+                node.info().name,
             ))
         }
         Circuit::Scatter(node) => {
@@ -320,7 +322,7 @@ pub fn expand_node(
                     .cloned()
                     .chain(node.info().shape.iter().cloned())
                     .collect(),
-                node.name_cloned(),
+                node.info().name,
             ))
         }
         Circuit::Concat(node) => {
@@ -375,7 +377,7 @@ pub fn expand_node(
                 inputs.clone()
             };
 
-            Concat::try_new(inputs, new_axis, node.name_cloned()).map(|x| x.rc())
+            Concat::try_new(inputs, new_axis, node.info().name).map(|x| x.rc())
         }
         Circuit::Add(node) => {
             let inputs: Vec<_> = inputs
@@ -387,7 +389,7 @@ pub fn expand_node(
                     if rank_diff > 0 && b_rank > 0 {
                         inp.unsqueeze(
                             (b_rank..rank_diff + b_rank).collect(),
-                            inp.name().map(|x| format!("{} unsq_b_add", x)),
+                            inp.info().name.map(|x| format!("{} unsq_b_add", x).into()),
                         )
                         .unwrap()
                         .rc()
@@ -435,7 +437,7 @@ pub fn expand_node(
                 inputs.clone()
             };
 
-            Add::try_new(inputs, node.name_cloned()).map(|x| x.rc())
+            Add::try_new(inputs, node.info().name).map(|x| x.rc())
         }
         Circuit::GeneralFunction(node) => {
             // TODO: symbolic? (annoying to handle, would require changes to spec)
@@ -455,7 +457,7 @@ pub fn expand_node(
                     }
                 })?,
                 node.spec.clone(),
-                node.name_cloned(),
+                node.info().name,
             )
             .map(|x| x.rc())
         }
@@ -525,9 +527,9 @@ pub fn expand_node(
                     )
                 })?;
 
-            Einsum::try_new(new_args, out_axes, node.name_cloned()).map(|x| x.rc())
+            Einsum::try_new(new_args, out_axes, node.info().name).map(|x| x.rc())
         }
-        Circuit::Tag(node) => Ok(Tag::nrc(inputs[0].clone(), node.uuid, node.name_cloned())),
+        Circuit::Tag(node) => Ok(Tag::nrc(inputs[0].clone(), node.uuid, node.info().name)),
         Circuit::Cumulant(node) => {
             // For now, only supports expansion in easy cases
             if zip(inputs, &node.nodes)
@@ -536,7 +538,7 @@ pub fn expand_node(
             {
                 bail!(ExpandError::CumulantRankChanged { circuit })
             }
-            Ok(Cumulant::nrc(inputs.clone(), node.name_cloned()))
+            Ok(Cumulant::nrc(inputs.clone(), node.info().name))
         }
         Circuit::Module(node) => {
             assert_eq!(inputs.len(), node.children().count());
@@ -552,7 +554,7 @@ pub fn expand_node(
                         circuit: spec_circuit,
                         arg_specs: vec![],
                     },
-                    node.name_cloned(),
+                    node.info().name,
                 ));
             }
 
@@ -645,13 +647,16 @@ pub fn expand_node(
                                     .cloned()
                                     .collect(),
                                 arg_spec.symbol.uuid,
-                                arg_spec.symbol.name_cloned(),
+                                arg_spec.symbol.info().name,
                             );
                             let node = if this_current_batch_rank > 0 && this_sym_extra_rank > 0 {
                                 let old_sym_rank = arg_spec.symbol.rank();
                                 let node_rank = node.rank();
                                 let up_to_old_sym_rank = node_rank - old_sym_rank;
-                                let shift_name = node.name().map(|x| format!("{} shift_batch", x));
+                                let shift_name = node
+                                    .info()
+                                    .name
+                                    .map(|x| format!("{} shift_batch", x).into());
                                 let spec = RearrangeSpec::new_permute(
                                     (this_sym_extra_rank..up_to_old_sym_rank)
                                         .chain(0..this_sym_extra_rank)
@@ -704,7 +709,7 @@ pub fn expand_node(
                 arg_specs,
             };
 
-            let out = Module::try_new(new_nodes, spec, node.name_cloned())?.rc();
+            let out = Module::try_new(new_nodes, spec, node.info().name)?.rc();
 
             // unshift dims
             let final_out = if current_batch_rank > 0 && rank_to_push_into_spec_circuit > 0 {
@@ -738,8 +743,8 @@ pub fn expand_node(
                 .unwrap();
                 assert!(!spec.is_identity()); // if statement should be checking for identity
 
-                let shift_name = out.name().map(|x| format!("{} shift_batch", x));
-                let out_name = out.name_cloned();
+                let shift_name = out.info().name.map(|x| format!("{} shift_batch", x).into());
+                let out_name = out.info().name;
                 Rearrange::nrc(out.rename(shift_name), spec, out_name)
             } else {
                 out
