@@ -12,6 +12,7 @@ use pyo3::{
 use rand::Rng;
 use regex::Regex;
 use rr_util::{
+    name::Name,
     py_types::{assert_tensors_close, ExtraPySelfOps, PyShape, Tensor, PY_UTILS},
     pycall, python_error_exception, simple_from,
     tensor_util::{
@@ -41,8 +42,8 @@ macro_rules! gf_gen {
 
         $(
         #[pyfunction]
-        pub fn $name(circuit: CircuitRc, name: Option<String>) -> Result<GeneralFunction> {
-            GeneralFunction::new_by_name(vec![circuit], stringify!($name).to_owned(), name)
+        pub fn $name(circuit: CircuitRc, name: Option<Name>) -> Result<GeneralFunction> {
+            GeneralFunction::new_by_name(vec![circuit], stringify!($name).into(), name)
         }
         )*
 
@@ -78,16 +79,20 @@ gf_gen!(
     (min, 0, 1),
     (max, 0, 1),
     (last_dim_size, 0, 1),
+    (abs, 0),
+    (exp, 0),
+    (log, 0),
+    (logit, 0),
 );
 
-static SPECS: GILLazy<HashMap<String, GeneralFunctionSpec>> = GILLazy::new(|| {
+static SPECS: GILLazy<HashMap<Name, GeneralFunctionSpec>> = GILLazy::new(|| {
     BASIC_SPEC_ITEMS
         .iter()
         .cloned()
         .map(|(name, num_non_batchable_output_dims, removed_from_end)| {
-            let name = name.to_owned();
+            let name = name.into();
             (
-                name.clone(),
+                name,
                 GeneralFunctionSimpleSpec {
                     name,
                     num_non_batchable_output_dims,
@@ -110,7 +115,7 @@ pub trait SpecTrait: fmt::Debug + ToPyObject {
     fn get_shape_info(&self, shapes: &[Shape]) -> Result<GeneralFunctionShapeInfo>;
     // fn has_jacobian(&self) -> Result<bool>;
     // fn get_jacobians(&self, func: &GeneralFunction) -> Result<Option<Vec<Circuit>>>;
-    fn name(&self) -> String;
+    fn name(&self) -> &'static str;
     fn is_official(&self) -> bool {
         false
     }
@@ -160,7 +165,7 @@ impl GeneralFunctionShapeInfo {
 #[derive(Debug, Clone)]
 pub struct GeneralFunctionSimpleSpec {
     #[pyo3(get)]
-    name: String,
+    name: Name,
     #[pyo3(get)]
     num_non_batchable_output_dims: u8,
     #[pyo3(get)]
@@ -184,7 +189,7 @@ pub fn get_shape_info_simple(
     removed_from_end: Option<u8>,
 ) -> Result<GeneralFunctionShapeInfo> {
     GeneralFunctionSimpleSpec {
-        name: "".to_owned(),
+        name: "".into(),
         num_non_batchable_output_dims: num_non_batchable_output_dims.unwrap_or(0),
         removed_from_end: removed_from_end.unwrap_or(0),
     }
@@ -194,7 +199,7 @@ pub fn get_shape_info_simple(
 #[pymethods]
 impl GeneralFunctionSimpleSpec {
     fn get_function(&self) -> PyObject {
-        PY_UTILS.generalfunctions[&self.name].clone()
+        PY_UTILS.generalfunctions[&self.name.string()].clone()
     }
 }
 
@@ -252,14 +257,14 @@ impl SpecTrait for GeneralFunctionSimpleSpec {
     // }
     // fn get_jacobians(&self, func: &GeneralFunction) -> Result<Option<Vec<Circuit>>> {}
 
-    fn name(&self) -> String {
-        self.name.clone()
+    fn name(&self) -> &'static str {
+        self.name.into()
     }
     fn is_official(&self) -> bool {
         true
     }
     fn serialize(&self) -> Result<Option<String>> {
-        Ok(Some(self.name.clone()))
+        Ok(Some(self.name().into()))
     }
 }
 
@@ -395,19 +400,20 @@ impl SpecTrait for GeneralFunctionIndexSpec {
         }
         Some(Ok(device_dtypes[0]))
     }
-    fn name(&self) -> String {
-        format!(
+    fn name(&self) -> &'static str {
+        Name::new(&format!(
             "gen_index_at_{}{}{}",
             self.index_dim,
             if self.batch_x { "_batch_x" } else { "" },
             if self.check_index_ints { "_c" } else { "" }
-        )
+        ))
+        .into()
     }
     fn is_official(&self) -> bool {
         true
     }
     fn serialize(&self) -> Result<Option<String>> {
-        Ok(Some(self.name()))
+        Ok(Some(self.name().into()))
     }
 }
 
@@ -482,14 +488,15 @@ impl SpecTrait for GeneralFunctionSetDDSpec {
     // }
     // fn get_jacobians(&self, func: &GeneralFunction) -> Result<Option<Vec<Circuit>>> {}
 
-    fn name(&self) -> String {
-        format!(
+    fn name(&self) -> &'static str {
+        Name::new(&format!(
             "cast_{:?}_{:?}",
             self.input_required_compatibility, self.output
-        )
+        ))
+        .into()
     }
     fn serialize(&self) -> Result<Option<String>> {
-        Ok(Some(self.name()))
+        Ok(Some(self.name().to_owned()))
     }
 }
 
@@ -497,7 +504,7 @@ impl SpecTrait for GeneralFunctionSetDDSpec {
 pub struct PyWrap {
     ob: PyObject,
     hash: HashBytes,
-    name: String,
+    name: Name,
     path: Option<String>,
 }
 
@@ -562,7 +569,7 @@ impl<'source> pyo3::FromPyObject<'source> for Box<PyWrap> {
         }
 
         Ok(Box::new(PyWrap {
-            name,
+            name: name.into(),
             path,
             ob,
             hash: *hasher.finalize().as_bytes(),
@@ -636,8 +643,8 @@ impl SpecTrait for Box<PyWrap> {
         }
         return resulty.unwrap().map(|x| Ok(x));
     }
-    fn name(&self) -> String {
-        self.name.clone()
+    fn name(&self) -> &'static str {
+        self.name.into()
     }
     fn serialize(&self) -> Result<Option<String>> {
         match &self.path {
@@ -699,7 +706,7 @@ impl SpecTrait for GeneralFunctionSpec {
     }
     // fn has_jacobian(&self) -> Result<bool>;
     // fn get_jacobians(&self, func: &GeneralFunction) -> Result<Option<Vec<Circuit>>>;
-    fn name(&self) -> String {
+    fn name(&self) -> &'static str {
         self.as_trait_obj().name()
     }
     fn is_official(&self) -> bool {
@@ -718,7 +725,6 @@ pub struct GeneralFunction {
     #[pyo3(get)]
     pub spec: GeneralFunctionSpec,
     info: CachedCircuitInfo,
-    name: Option<String>,
     #[pyo3(get)]
     pub num_non_batchable_output_dims: u8,
     pub input_batchability: Vec<bool>,
@@ -729,7 +735,7 @@ impl GeneralFunction {
     pub fn try_new(
         nodes: Vec<CircuitRc>,
         spec: GeneralFunctionSpec,
-        name: Option<String>,
+        name: Option<Name>,
     ) -> Result<Self> {
         let shapes = nodes
             .iter()
@@ -754,12 +760,12 @@ impl GeneralFunction {
         let mut out = Self {
             nodes,
             spec,
-            name,
             info: Default::default(),
             num_non_batchable_output_dims,
             input_batchability,
         };
         out.info.shape = shape; // set shape so we only compute once
+        out.info.name = name;
         out.initial_init_info()
     }
 
@@ -810,7 +816,7 @@ impl CircuitNode for GeneralFunction {
                 .map(move |(i, circ)| f(i, circ.clone()))
                 .collect::<Result<Vec<_>, _>>()?,
             self.spec.clone(),
-            self.name.clone(),
+            self.info().name,
         )
     }
 
@@ -840,21 +846,22 @@ impl CircuitNode for GeneralFunction {
 impl CircuitNodeAutoName for GeneralFunction {
     const PRIORITY: OperatorPriority = OperatorPriority::Function {};
 
-    fn auto_name(&self) -> Option<String> {
+    fn auto_name(&self) -> Option<Name> {
         // Never any parenthesis, so we don't even care checking if we need to add some
         // but check out children_names_with_maybe_paren if you change the syntax of the general function
-        if self.children().any(|x| x.name().is_none()) {
+        if self.children().any(|x| x.info().name.is_none()) {
             None
         } else {
             Some(
-                self.spec.name().clone()
+                (self.spec.name().to_owned()
                     + "("
                     + &self
                         .children()
-                        .map(|x| Self::shorten_child_name(x.name().unwrap()))
+                        .map(|x| Self::shorten_child_name(x.info().name.unwrap().str()))
                         .collect::<Vec<String>>()
                         .join(", ")
-                    + ")",
+                    + ")")
+                    .into(),
             )
         }
     }
@@ -867,7 +874,7 @@ impl GeneralFunction {
     fn new_py(
         nodes: Vec<CircuitRc>,
         spec: GeneralFunctionSpec,
-        name: Option<String>,
+        name: Option<Name>,
     ) -> PyResult<PyClassInitializer<GeneralFunction>> {
         let out = GeneralFunction::try_new(nodes, spec, name)?;
 
@@ -879,27 +886,23 @@ impl GeneralFunction {
     pub fn new_from_parse(
         nodes: Vec<CircuitRc>,
         parse_string: String,
-        name: Option<String>,
+        name: Option<Name>,
     ) -> Result<Self> {
         if parse_string.contains(" at ") {
             let mut split = parse_string.splitn(2, " at ");
             let name = convert_op_str_to_format_err(split.next(), &parse_string)?;
             let path = convert_op_str_to_format_err(split.next(), &parse_string)?;
 
-            Self::new_by_path(nodes, path, Some(name))
+            Self::new_by_path(nodes, path, Some(name.into()))
         } else {
-            Self::new_by_name(nodes, parse_string, name)
+            Self::new_by_name(nodes, parse_string.into(), name)
         }
     }
 
     #[staticmethod]
     #[args(nodes = "*", spec_name, name = "None")]
-    pub fn new_by_name(
-        nodes: Vec<CircuitRc>,
-        spec_name: String,
-        name: Option<String>,
-    ) -> Result<Self> {
-        Self::new_by_name_op(nodes, spec_name.clone(), name)?
+    pub fn new_by_name(nodes: Vec<CircuitRc>, spec_name: Name, name: Option<Name>) -> Result<Self> {
+        Self::new_by_name_op(nodes, spec_name, name)?
             .ok_or(ConstructError::UnknownGeneralFunction { spec_name }.into())
     }
 
@@ -907,8 +910,8 @@ impl GeneralFunction {
     #[args(nodes = "*", spec_name, name = "None")]
     pub fn new_by_name_op(
         nodes: Vec<CircuitRc>,
-        spec_name: String,
-        name: Option<String>,
+        spec_name: Name,
+        name: Option<Name>,
     ) -> Result<Option<Self>> {
         // parse index case
         static RE: Lazy<Regex> =
@@ -939,7 +942,7 @@ impl GeneralFunction {
 
     #[staticmethod]
     #[args(nodes = "*", path, name = "None")]
-    pub fn new_by_path(nodes: Vec<CircuitRc>, path: String, name: Option<String>) -> Result<Self> {
+    pub fn new_by_path(nodes: Vec<CircuitRc>, path: String, name: Option<Name>) -> Result<Self> {
         let mut split = path.splitn(2, ":");
         let file_path = convert_op_str_to_format_err(split.next(), &path)?;
         let spec_name = convert_op_str_to_format_err(split.next(), &path)?;
@@ -977,7 +980,7 @@ impl GeneralFunction {
         index_dim: i64,
         batch_x: bool,
         check_index_ints: bool,
-        name: Option<String>,
+        name: Option<Name>,
     ) -> Result<Self> {
         let spec = GeneralFunctionIndexSpec {
             index_dim,
@@ -997,7 +1000,7 @@ impl GeneralFunction {
         x: CircuitRc,
         input_required_compatibility: TorchDeviceDtypeOp,
         output: TorchDeviceDtypeOp,
-        name: Option<String>,
+        name: Option<Name>,
     ) -> Result<Self> {
         let spec = GeneralFunctionSetDDSpec {
             input_required_compatibility,
@@ -1420,7 +1423,7 @@ fn test_simple_general_function() -> Result<()> {
     for num_non_batchable_output_dims in 0..2 {
         for removed_from_end in 0..2 {
             let spec = GeneralFunctionSimpleSpec {
-                name: "".to_owned(),
+                name: "".into(),
                 num_non_batchable_output_dims,
                 removed_from_end,
             }
