@@ -1313,10 +1313,16 @@ impl Not for CircuitFlags {
     }
 }
 
-/// don't want to print hash with Debug, for now just print shape
+/// don't want to print hash with Debug; print selected fields
 impl Debug for CachedCircuitInfo {
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
-        write!(f, "{:?}", self.shape)
+        write!(
+            f,
+            "{} {:?} {:?}",
+            self.name.map_or("", |n| n.str()),
+            self.shape,
+            self.device_dtype,
+        )
     }
 }
 
