@@ -13,6 +13,7 @@ use rr_util::{
     cached_lambda, cached_method,
     caching::FastUnboundedCache,
     eq_by_big_hash, impl_eq_by_big_hash,
+    name::Name,
     py_types::{use_rust_comp, PyCallable},
     pycall, python_error_exception,
     rearrange_spec::RearrangeSpec,
@@ -238,7 +239,7 @@ impl ModuleSpec {
             )
             .context("rank overflow in module expand")?;
 
-            let rep_name = out.name().map(|x| format!("{} rep_batch", x));
+            let rep_name: Option<Name> = out.info().name.map(|x| format!("{} rep_batch", x).into());
             Rearrange::nrc(out, spec, rep_name)
         } else {
             out
@@ -265,12 +266,15 @@ impl ModuleSpec {
                 Symbol::nrc(
                     s.clone(),
                     self.get_spec_circuit_uuid(),
-                    Some(format!(
-                        "{}_internal_expand_shape_{}_arg{}",
-                        uuid,
-                        self.circuit.name().unwrap_or(""),
-                        i
-                    )),
+                    Some(
+                        format!(
+                            "{}_internal_expand_shape_{}_arg{}",
+                            uuid,
+                            Name::str_maybe_empty(self.circuit.info().name),
+                            i
+                        )
+                        .into(),
+                    ),
                 )
             })
             .collect();
@@ -298,7 +302,7 @@ impl ModuleSpec {
             .map(|x| [x.clone()].into_iter().collect());
         let syms = sym_single_map
             .as_ref()
-            .unwrap_or(&self.circuit.info().free_symbols);
+            .unwrap_or(self.circuit.info().get_raw_free_symbols());
 
         for sym_here in syms {
             if let Some(&my_sym) = my_idents.get(&sym_here.ident()) {
@@ -378,19 +382,20 @@ impl ModuleSpec {
             // (NOTE: these free symbols are now the *expanded* symbols from expand_shape)
             if !x.is_symbol()
                 && x.info()
-                    .free_symbols
+                    .get_raw_free_symbols()
                     .iter()
                     .any(|x| node_mapping.contains_key(&x.info().hash))
             {
-                if let Some(n) = x.name() {
+                if let Some(n) = x.info().name {
                     if n.contains(&clear_names_str) {
                         let name = x.get_autoname().unwrap_or(None);
                         return Some(x.rename(name));
                     } else if name_prefix.is_some() || name_suffix.is_some() {
                         let name = Some(
-                            name_prefix.as_deref().unwrap_or("").to_owned()
-                                + n
-                                + name_suffix.as_deref().unwrap_or(""),
+                            (name_prefix.as_deref().unwrap_or("").to_owned()
+                                + n.into()
+                                + name_suffix.as_deref().unwrap_or(""))
+                            .into(),
                         );
                         return Some(x.rename(name));
                     }
@@ -411,8 +416,6 @@ impl ModuleSpec {
                 concat!(
                     "replacing symbols with nodes",
                     " failed in ModuleSpec substitute\n",
-                    "This is typically caused by multiple dtypes ",
-                    "(see root cause below)\n",
                     "module_spec={:?}\nnodes={:?}"
                 ),
                 self, nodes
@@ -477,7 +480,7 @@ impl ModuleSpec {
             let new = if let Some(m) = circ.as_module() {
                 let mut removed: BTreeSet<_> = symbols
                     .iter()
-                    .filter(|x| m.info().free_symbols.contains(*x))
+                    .filter(|x| m.info().get_raw_free_symbols().contains(*x))
                     .cloned()
                     .collect();
                 for bound in &m.spec.arg_specs {
@@ -533,7 +536,7 @@ pub fn get_free_symbols(circuit: &Circuit) -> FreeSymbolsWrapper<'_> {
     if let Some(sym) = circuit.as_symbol() {
         FreeSymbolsWrapper::Sym([sym.clone()].into_iter().collect())
     } else {
-        FreeSymbolsWrapper::Other(&circuit.info().free_symbols)
+        FreeSymbolsWrapper::Other(&circuit.info().get_raw_free_symbols())
     }
 }
 
@@ -581,19 +584,23 @@ impl ModuleSpec {
 
     pub fn check_unique_arg_names(&self) -> Result<()> {
         // TODO: maybe cache me (as needed)!
-        if self.arg_specs.iter().any(|x| x.symbol.name().is_none()) {
+        if self
+            .arg_specs
+            .iter()
+            .any(|x| x.symbol.info().name.is_none())
+        {
             bail!(ConstructError::ModuleSomeArgsNamedNone {
                 symbols_named_none: self
                     .arg_specs
                     .iter()
-                    .filter_map(|x| x.symbol.name().is_none().then(|| x.symbol.clone()))
+                    .filter_map(|x| x.symbol.info().name.is_none().then(|| x.symbol.clone()))
                     .collect()
             })
         }
-        let names: Vec<_> = self
+        let names: Vec<Name> = self
             .arg_specs
             .iter()
-            .map(|x| x.symbol.name().unwrap())
+            .map(|x| x.symbol.info().name.unwrap())
             .collect();
         if !is_unique(&names) {
             bail!(ConstructError::ModuleArgsDupNames {
@@ -651,7 +658,7 @@ impl ModuleSpec {
                     arg_spec.symbol = Symbol::new(
                         circuit.info().shape.clone(),
                         arg_spec.symbol.uuid,
-                        arg_spec.symbol.name_cloned().or(circuit.name_cloned()),
+                        arg_spec.symbol.info().name.or(circuit.info().name),
                     );
                     real_arg_specs[i] = Some(arg_spec);
                     return (Some(real_arg_specs[i].as_ref().unwrap().symbol.crc()), true);
@@ -695,7 +702,7 @@ impl ModuleSpec {
     pub fn resize(&self, shapes: Vec<Shape>) -> Result<Self> {
         let arg_specs: Vec<ModuleArgSpec> = zip(&self.arg_specs, shapes)
             .map(|(arg_spec, shape)| ModuleArgSpec {
-                symbol: Symbol::new(shape, arg_spec.symbol.uuid, arg_spec.symbol.name_cloned()),
+                symbol: Symbol::new(shape, arg_spec.symbol.uuid, arg_spec.symbol.info().name),
                 ..arg_spec.clone()
             })
             .collect();
@@ -739,11 +746,12 @@ impl ModuleSpec {
         }
 
         self.map_on_replaced_path_unwrap(|x| {
-            if let Some(n) = x.name() {
+            if let Some(n) = x.info().name {
                 let name = Some(
-                    prefix.as_deref().unwrap_or("").to_owned()
-                        + n
-                        + suffix.as_deref().unwrap_or(""),
+                    (prefix.as_deref().unwrap_or("").to_owned()
+                        + n.into()
+                        + suffix.as_deref().unwrap_or(""))
+                    .into(),
                 );
                 x.rename(name)
             } else {
@@ -826,10 +834,7 @@ impl ModuleArgSpec {
         ban_non_symbolic_size_expand: bool,
     ) -> Self {
         Self {
-            symbol: Symbol::new_with_random_uuid(
-                circuit.info().shape.clone(),
-                circuit.name_cloned(),
-            ),
+            symbol: Symbol::new_with_random_uuid(circuit.info().shape.clone(), circuit.info().name),
             batchable,
             expandable,
             ban_non_symbolic_size_expand,
@@ -862,13 +867,11 @@ pub struct Module {
     pub spec: ModuleSpec,
     substitute_cached_info: Box<CachedCircuitInfo>, // box for size
     info: CachedCircuitInfo,
-    #[pyo3(get)]
-    name: Option<String>,
 }
 
 impl Module {
     #[apply(new_rc_unwrap)]
-    pub fn try_new(nodes: Vec<CircuitRc>, spec: ModuleSpec, name: Option<String>) -> Result<Self> {
+    pub fn try_new(nodes: Vec<CircuitRc>, spec: ModuleSpec, name: Option<Name>) -> Result<Self> {
         // we need to check this here because we substitute modules below where other checks run
         spec.check_no_free_bound_by_nested(&nodes).with_context(|| {
             const MSG: &str = "(free sym bound by nested)";
@@ -897,24 +900,24 @@ impl Module {
         for n in &nodes {
             free_syms.extend(get_free_symbols(&**n).iter().cloned());
         }
-        substitute_cached_info.free_symbols = free_syms;
+        substitute_cached_info.free_symbols = Some(Arc::new(free_syms));
 
         let substitute_cached_info = Box::new(substitute_cached_info);
 
-        let out = Self {
+        let mut out = Self {
             nodes,
             spec,
-            name,
             substitute_cached_info,
             info: Default::default(),
         };
+        out.info.name = name;
         out.initial_init_info()
     }
 
     pub fn new_kwargs(
-        kwargs: &HashMap<String, CircuitRc>,
+        kwargs: &HashMap<Name, CircuitRc>,
         spec: ModuleSpec,
-        name: Option<String>,
+        name: Option<Name>,
     ) -> Result<Self> {
         let mut nodes: Vec<_> = vec![None; spec.arg_specs.len()];
         spec.check_unique_arg_names()?;
@@ -922,18 +925,18 @@ impl Module {
             match spec
                 .arg_specs
                 .iter()
-                .position(|x| x.symbol.name().expect("check_unique_arg_names above") == k)
+                .position(|x| x.symbol.info().name.expect("check_unique_arg_names above") == *k)
             {
                 Some(i) => {
                     nodes[i] = Some(v.clone());
                 }
                 None => {
                     bail!(ConstructError::ModuleUnknownArgument {
-                        argument: k.clone(),
+                        argument: *k,
                         all_module_inputs: spec
                             .arg_specs
                             .iter()
-                            .map(|x| x.symbol.name_cloned().unwrap())
+                            .map(|x| x.symbol.info().name.unwrap())
                             .collect()
                     })
                 }
@@ -946,7 +949,8 @@ impl Module {
                     .zip(spec.arg_specs)
                     .filter_map(|(n, arg_spec)| n.is_none().then(|| arg_spec
                         .symbol
-                        .name_cloned()
+                        .info()
+                        .name
                         .expect("check_unique_arg_names above")))
                     .collect()
             });
@@ -965,7 +969,7 @@ impl Module {
         let out_named = propagate_named_axes(
             stripped,
             (0..self.info().rank())
-                .map(|i| (i as u8, i.to_string()))
+                .map(|i| (i as u8, i.to_string().into()))
                 .collect(),
             true,
         );
@@ -976,12 +980,9 @@ impl Module {
             .collect();
         visit_circuit_unwrap(out_named, |x| {
             if let Some(sym) = x.as_symbol() {
-                if let Some(i) = self
-                    .spec
-                    .arg_specs
-                    .iter()
-                    .position(|x| x.symbol.uuid == sym.uuid && x.symbol.name() == sym.name())
-                {
+                if let Some(i) = self.spec.arg_specs.iter().position(|x| {
+                    x.symbol.uuid == sym.uuid && x.symbol.info().name == sym.info().name
+                }) {
                     for (k, v) in &sym.info().named_axes {
                         result[i][*k as usize] = Some(v.parse::<usize>().unwrap());
                     }
@@ -1002,7 +1003,9 @@ impl CircuitNodeHashItems for Module {
 
 impl CircuitNodeSetNonHashInfo for Module {
     fn set_non_hash_info(&mut self) -> Result<()> {
+        let old_name = self.info().name;
         *self.info_mut() = (*self.substitute_cached_info).clone();
+        self.info_mut().name = old_name;
         Ok(())
     }
 }
@@ -1051,7 +1054,7 @@ impl CircuitNode for Module {
                 circuit: new_spec_circ,
                 arg_specs: self.spec.arg_specs.clone(),
             },
-            self.name.clone(),
+            self.info().name,
         )
     }
 
@@ -1066,7 +1069,7 @@ impl CircuitNode for Module {
             .enumerate()
             .map(|(i, x)| f(i, x))
             .collect::<Result<_>>()?;
-        Self::try_new(nodes, self.spec.clone(), self.name_cloned())
+        Self::try_new(nodes, self.spec.clone(), self.info().name)
     }
 
     fn eval_tensors(
@@ -1083,20 +1086,21 @@ impl CircuitNode for Module {
 impl CircuitNodeAutoName for Module {
     const PRIORITY: OperatorPriority = OperatorPriority::Function {};
 
-    fn auto_name(&self) -> Option<String> {
-        if self.children().any(|x| x.name().is_none()) {
+    fn auto_name(&self) -> Option<Name> {
+        if self.children().any(|x| x.info().name.is_none()) {
             None
         } else {
             Some(
-                self.spec.circuit.name_cloned().unwrap()
+                (self.spec.circuit.info().name.unwrap().string()
                     + "("
                     + &self
                         .nodes
                         .iter()
-                        .map(|node| node.name_cloned().unwrap())
+                        .map(|node| node.info().name.unwrap().into())
                         .collect::<Vec<String>>()
                         .join(", ")
-                    + ")",
+                    + ")")
+                    .into(),
             )
         }
     }
@@ -1108,15 +1112,15 @@ impl Module {
     #[args(name = "None", kwargs = "**")]
     fn new_py(
         spec: ModuleSpec,
-        name: Option<String>,
-        kwargs: Option<HashMap<String, CircuitRc>>,
+        name: Option<Name>,
+        kwargs: Option<HashMap<Name, CircuitRc>>,
     ) -> PyResult<PyClassInitializer<Module>> {
         Ok(Module::new_kwargs(&kwargs.unwrap_or_else(HashMap::default), spec, name)?.into_init())
     }
 
     #[staticmethod]
     #[args(nodes = "*", name = "None")]
-    fn new_flat(spec: ModuleSpec, nodes: Vec<CircuitRc>, name: Option<String>) -> Result<Self> {
+    fn new_flat(spec: ModuleSpec, nodes: Vec<CircuitRc>, name: Option<Name>) -> Result<Self> {
         Self::try_new(nodes, spec, name)
     }
 
@@ -1148,7 +1152,7 @@ impl Module {
         Ok(Self::new(
             self.nodes.clone(),
             self.spec.rename_on_replaced_path(prefix, suffix),
-            self.name_cloned(),
+            self.info().name,
         ))
     }
 
@@ -1156,7 +1160,7 @@ impl Module {
         Ok(Self::new(
             self.nodes.clone(),
             self.spec.py_map_on_replaced_path(f)?,
-            self.name_cloned(),
+            self.info().name,
         ))
     }
 
@@ -1196,7 +1200,7 @@ impl Module {
                         .collect(),
                 )
                 .expect("constructor should ensure this works"),
-            self.name_cloned(),
+            self.info().name,
         ))
     }
 
@@ -1207,7 +1211,7 @@ impl Module {
             self.spec
                 .resize(self.nodes.iter().map(|x| x.shape().clone()).collect())
                 .expect("constructor should ensure this works"),
-            self.name_cloned(),
+            self.info().name,
         )
     }
 
@@ -1238,7 +1242,7 @@ impl Module {
 
 impl Module {
     pub fn substitute_self_name_prefix(&self) -> CircuitRc {
-        self.substitute(self.name_cloned().map(|x| x + "."), None)
+        self.substitute(self.info().name.map(|x| x.string() + "."), None)
     }
 
     pub fn substitute(
@@ -1249,7 +1253,7 @@ impl Module {
         self.spec
             .substitute(&self.nodes, name_prefix, name_suffix)
             .expect("constructor should ensure this works")
-            .rename(self.name_cloned())
+            .rename(self.info().name)
     }
 
     fn get_prefix(
@@ -1267,7 +1271,7 @@ impl Module {
                 )
             }
 
-            Ok(self.name_cloned().map(|x| x + "."))
+            Ok(self.info().name.map(|x| x.string() + "."))
         } else {
             Ok(name_prefix)
         }
@@ -1307,7 +1311,7 @@ impl AllModuleSubstituter {
             // Not clear if this would be slower or faster overall.
             .substitute(&nodes, None, None)
             .expect("module constructor should ensure this works")
-            .rename(m.name_cloned())
+            .rename(m.info().name)
         } else {
             circuit.map_children_unwrap_idxs(|i| new_children[i].clone())
         }
@@ -1379,7 +1383,7 @@ pub fn has_free_sym(circuit: &CircuitRc, sym: &Symbol) -> bool {
     if let Some(other) = circuit.as_symbol() {
         other == sym
     } else {
-        circuit.info().free_symbols.contains(sym)
+        circuit.info().get_raw_free_symbols().contains(sym)
     }
 }
 
@@ -1391,7 +1395,7 @@ pub fn is_intersecting_free_syms_iter<'a>(
         syms.into_iter().any(|x| x == other)
     } else {
         syms.into_iter()
-            .any(|x| circuit.info().free_symbols.contains(x))
+            .any(|x| circuit.info().get_raw_free_symbols().contains(x))
     }
 }
 
@@ -1399,7 +1403,7 @@ pub fn is_intersecting_free_syms(circuit: &CircuitRc, syms: &IndexSet<Symbol>) -
     if let Some(other) = circuit.as_symbol() {
         syms.contains(other)
     } else {
-        !circuit.info().free_symbols.is_disjoint(syms)
+        !circuit.info().get_raw_free_symbols().is_disjoint(syms)
     }
 }
 
