@@ -13,8 +13,8 @@ use num_bigint::BigUint;
 use once_cell::unsync::Lazy;
 use pyo3::{once_cell::GILLazy, prelude::*};
 use rr_util::{
-    fn_struct, py_types::MaybeNotSet, python_println, symbolic_size::SymbolicSizeProduct,
-    tensor_db::write_tensor_to_dir_tree, util::ALPHABET,
+    fn_struct, name::Name, py_types::MaybeNotSet, python_println,
+    symbolic_size::SymbolicSizeProduct, tensor_db::write_tensor_to_dir_tree, util::ALPHABET,
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 
@@ -126,7 +126,7 @@ pub struct PrintOptions {
     #[pyo3(get, set)]
     pub bijection: bool,
     #[pyo3(get)]
-    pub reference_circuits: HashMap<CircuitRc, String>, /* TODO: make a checking setter for me as needed */
+    pub reference_circuits: HashMap<CircuitRc, Name>, /* TODO: make a checking setter for me as needed */
     #[pyo3(get, set)]
     pub shape_only_when_necessary: bool,
     pub traversal: Option<OpaqueIterativeMatcherVal>,
@@ -209,7 +209,7 @@ impl PrintOptions {
         result: &mut String,
         seen_hashes: &mut HashMap<HashBytes, (String, String)>,
         runnning_serial_number: &mut usize,
-        disallow_name_ident: Option<&HashMap<String, bool>>,
+        disallow_name_ident: Option<&HashMap<Name, bool>>,
     ) -> Result<bool> {
         if let Some(ident) = self.reference_circuits.get(&circ.clone().rc()) {
             result.push_str(&self.string_escape(ident));
@@ -228,11 +228,15 @@ impl PrintOptions {
 
         // allow_name_ident is_some iff !self.force_use_serial_numbers
         let disallow_name = disallow_name_ident
-            .zip(circ.name())
-            .map(|(map, name)| map[name])
+            .zip(circ.info().name)
+            .map(|(map, name)| map[&name])
             .unwrap_or(true);
 
-        let (repeat_string, line_prefix) = match (repeat_info, disallow_name, circ.name()) {
+        let (repeat_string, line_prefix) = match (
+            repeat_info,
+            disallow_name,
+            circ.info().name.map(|z| z.str()),
+        ) {
             (Some(x), _, _) => x,
             (None, true, Some(name)) => {
                 let escaped = self.string_escape(name);
@@ -279,7 +283,7 @@ impl PrintOptions {
         circ: &Circuit,
         seen_hashes: &mut HashMap<HashBytes, (String, String)>,
         runnning_serial_number: &mut usize,
-        disallow_name_ident: Option<&HashMap<String, bool>>,
+        disallow_name_ident: Option<&HashMap<Name, bool>>,
         child_idx: usize,
     ) -> Option<String> {
         if let Some(m) = circ.as_module() {
@@ -400,7 +404,7 @@ impl PrintOptions {
     )]
     pub fn new(
         bijection: bool,
-        reference_circuits: HashMap<CircuitRc, String>,
+        reference_circuits: HashMap<CircuitRc, Name>,
         reference_circuits_by_name: Vec<CircuitRc>,
         shape_only_when_necessary: bool,
         traversal: Option<OpaqueIterativeMatcherVal>,
@@ -472,7 +476,7 @@ impl PrintOptions {
     pub fn evolve(
         &self,
         bijection: MaybeNotSet<bool>,
-        reference_circuits: MaybeNotSet<HashMap<CircuitRc, String>>,
+        reference_circuits: MaybeNotSet<HashMap<CircuitRc, Name>>,
         shape_only_when_necessary: MaybeNotSet<bool>,
         traversal: MaybeNotSet<Option<OpaqueIterativeMatcherVal>>,
         leaves_on_top: MaybeNotSet<bool>,
@@ -700,7 +704,7 @@ impl PrintOptions {
                             .context("failed to get general function spec serialize in print")?
                             .unwrap_or_else(|| format!("{} [NOT_SERIALIZABLE]", gf.spec.name()))
                     } else {
-                        gf.spec.name()
+                        gf.spec.name().to_owned()
                     }
                 }
                 Circuit::Symbol(sy) => {
@@ -756,18 +760,18 @@ impl PrintOptions {
         let mut seen_hashes = HashMap::default();
 
         let disallow_name_ident = if !self.force_use_serial_numbers {
-            let mut multiple_with_name: HashMap<String, bool> = HashMap::default();
+            let mut multiple_with_name: HashMap<Name, bool> = HashMap::default();
             let mut seen = HashSet::default();
 
             fn recurse(
                 circ: CircuitRc,
                 seen: &mut HashSet<HashBytes>,
-                multiple_with_name: &mut HashMap<String, bool>,
+                multiple_with_name: &mut HashMap<Name, bool>,
             ) {
                 if !seen.insert(circ.info().hash) {
                     return;
                 }
-                if let Some(name) = circ.name_cloned() {
+                if let Some(name) = circ.info().name {
                     use std::collections::hash_map::Entry::*;
                     match multiple_with_name.entry(name) {
                         Occupied(mut entry) => {
@@ -808,7 +812,7 @@ impl PrintOptions {
             selfy: &PrintOptions,
             is_last_child: &Vec<bool>,
             traversal: Option<OpaqueIterativeMatcherVal>,
-            disallow_name_ident: Option<&HashMap<String, bool>>,
+            disallow_name_ident: Option<&HashMap<Name, bool>>,
             parent_info: Option<(&Circuit, usize)>,
             mut node_comments: Vec<String>,
         ) -> Result<()> {
@@ -1191,8 +1195,9 @@ pub fn print_circuit_stats(circuit: &Circuit) {
     let mut result = String::new();
     result.push_str(
         &circuit
-            .name_cloned()
-            .map(|x| x + " ")
+            .info()
+            .name
+            .map(|x| x.string() + " ")
             .unwrap_or(" ".to_owned()),
     );
     result.push_str(&circuit.variant_string());
