@@ -659,7 +659,7 @@ pub fn generalfunction_pull_removable_axes(node: &GeneralFunction) -> Option<Cir
     let removable_axes_inp = get_removable_axes(&node.nodes[0]);
     let removable_batchable_inp: HashSet<usize> = removable_axes_inp
         .iter()
-        .filter(|x| **x < node.nodes[0].info().rank() - node.num_non_batchable_output_dims as usize)
+        .filter(|x| (**x + node.num_non_batchable_output_dims as usize) < node.info().rank())
         .cloned()
         .collect();
     if removable_batchable_inp.is_empty() {
@@ -670,13 +670,9 @@ pub fn generalfunction_pull_removable_axes(node: &GeneralFunction) -> Option<Cir
         node.spec.clone(),
         node.info().name,
     );
-    let removable_batchable_out: HashSet<usize> = removable_batchable_inp
-        .iter()
-        .map(|x| x + node.info().rank() - node.nodes[0].info().rank())
-        .collect();
     Some(Rearrange::nrc_elim_identity(
         new_generalfunction.rc(),
-        RearrangeSpec::unremove_axes(&removable_batchable_out, &node.info().shape),
+        RearrangeSpec::unremove_axes(&removable_batchable_inp, &node.info().shape),
         None,
     ))
 }
@@ -1223,8 +1219,8 @@ pub fn index_split_axes(node: &Index, top_axes: HashSet<usize>) -> Option<Index>
 pub fn rearrange_fuse(node: &Rearrange) -> Option<Rearrange> {
     node.node.as_rearrange().and_then(|inner| {
         RearrangeSpec::fuse(
-            &inner.conform_to_input_shape_spec(),
-            &node.conform_to_input_shape_spec(),
+            &inner.conform_to_input_shape_spec().canonicalize(true),
+            &node.conform_to_input_shape_spec().canonicalize(true),
         )
         .map(|spec| {
             Rearrange::new(
