@@ -1,6 +1,10 @@
-use std::simd::{u64x8, u8x8, SimdPartialEq};
+use std::{
+    cell::{Cell, UnsafeCell},
+    simd::{u64x8, u8x8, SimdPartialEq},
+};
 
 use anyhow::{bail, Result};
+use cached::proc_macro::cached;
 use circuit_base::{
     deep_map_op_context, deep_map_preorder_unwrap, deep_map_unwrap, on_circuit_names, prelude::*,
     visit_circuit_unwrap, Add, Index,
@@ -8,6 +12,7 @@ use circuit_base::{
 use itertools::Itertools;
 use macro_rules_attribute::apply;
 use num_bigint::BigUint;
+use once_cell::sync::Lazy;
 use pyo3::{exceptions::PyValueError, prelude::*};
 use rr_util::{
     python_error_exception, timed, timed_value,
@@ -40,7 +45,10 @@ use crate::{
 
 fn run_simp<T: CircuitNode + Into<Circuit> + Clone>(
     x: &'_ T,
-    fns: &[(&'static str, &dyn Fn(&'_ T) -> Option<CircuitRc>)],
+    fns: &[(
+        &'static str,
+        &(dyn Send + Sync + Fn(&'_ T) -> Option<CircuitRc>),
+    )],
     mut on_name: impl FnMut(&'static str),
 ) -> Option<CircuitRc> {
     for (name, f) in fns {
@@ -79,10 +87,10 @@ macro_rules! l_wrap {
 macro_rules! mk_simp_fns {
     ($($circ_ty:ident,)*) => (
         paste::paste! {
-            #[derive(Default)]
+            #[derive(Default,Clone)]
             struct SimpFns {
                 $(
-                    [<$circ_ty:snake>] : Vec<(&'static str, &'static dyn Fn(&circuit_base::$circ_ty) -> Option<CircuitRc>)>,
+                    [<$circ_ty:snake>] : Vec<(&'static str, &'static (dyn Send+Sync+Fn(&circuit_base::$circ_ty) -> Option<CircuitRc>))>,
                 )*
             }
 
@@ -137,10 +145,13 @@ macro_rules! mk_simp_fns {
         }
     )
 }
-
+fn all_simp_fns() -> &'static SimpFns {
+    &*ALL_SIMP_FNS_CACHE
+}
 on_circuit_names!(mk_simp_fns);
+static ALL_SIMP_FNS_CACHE: Lazy<SimpFns> = Lazy::new(all_simp_fns_raw);
 
-fn all_simp_fns() -> SimpFns {
+fn all_simp_fns_raw() -> SimpFns {
     SimpFns {
         add: vec![
             l_wrap!(remove_add_few_input),
@@ -211,8 +222,20 @@ fn all_simp_fns() -> SimpFns {
 }
 
 #[pyclass]
-#[derive(Debug, Clone, Eq, PartialEq)]
-pub struct SimpFnSubset(HashMap<String, bool>);
+#[derive(Clone)]
+pub struct SimpFnSubset(HashMap<String, bool>, SimpFns);
+
+impl PartialEq for SimpFnSubset {
+    fn eq(&self, other: &Self) -> bool {
+        self.0 == other.0
+    }
+}
+impl Eq for SimpFnSubset {}
+impl std::fmt::Debug for SimpFnSubset {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        self.0.fmt(f)
+    }
+}
 
 impl Default for SimpFnSubset {
     /// default to python simp, see compiler_default for compiler simplification
@@ -254,9 +277,8 @@ impl Default for SimpFnSubset {
 impl SimpFnSubset {
     #[staticmethod]
     pub fn all_names() -> Vec<&'static str> {
-        let fns = all_simp_fns();
         let mut out = Vec::new();
-        simp_fns_for_all!(&fns; fns_sub => out.extend(fns_sub.iter().map(|(s, _)| s)));
+        simp_fns_for_all!(all_simp_fns(); fns_sub => out.extend(fns_sub.iter().map(|(s, _)| s)));
         out
     }
 
@@ -271,7 +293,8 @@ impl SimpFnSubset {
             .into_iter()
             .map(|x| (x.to_owned(), include))
             .collect();
-        Self(mapping)
+        let fns = Self::get_simp_fns(&mapping);
+        Self(mapping, fns)
     }
 
     #[staticmethod]
@@ -347,7 +370,7 @@ impl SimpFnSubset {
         for k in names {
             new.insert(k, include).expect("should have this key");
         }
-        Ok(Self(new))
+        Ok(Self(new.clone(), Self::get_simp_fns(&new)))
     }
 
     #[args(set_vals = "**")]
@@ -360,7 +383,7 @@ impl SimpFnSubset {
         self.check_names(set_vals.keys().cloned().collect())?;
         let mut new = self.0.clone();
         new.extend(set_vals.into_iter().filter_map(|(k, v)| v.map(|v| (k, v))));
-        Ok(Self(new))
+        Ok(Self(new.clone(), Self::get_simp_fns(&new)))
     }
 
     pub fn exclude(&self, names: Vec<String>) -> Result<Self> {
@@ -382,13 +405,12 @@ impl SimpFnSubset {
 }
 
 impl SimpFnSubset {
-    fn simp_fns(&self) -> SimpFns {
-        map_simp_fns!(all_simp_fns(); fns_sub => fns_sub.into_iter().filter(|(s, _)| self.0[*s]).collect())
+    fn get_simp_fns(mapping: &HashMap<String, bool>) -> SimpFns {
+        map_simp_fns!(all_simp_fns(); fns_sub => fns_sub.iter().cloned().filter(|(s, _)| mapping[*s]).collect())
     }
-
     pub fn simp_step(&self, circuit: &Circuit) -> Option<CircuitRc> {
         match_circ_with_simp_fns!(
-            (circuit, self.simp_fns());
+            (circuit, &self.1);
             (node, fns_sub) => run_simp(node, &fns_sub, |_| ())
         )
     }
@@ -408,7 +430,7 @@ impl SimpFnSubset {
         };
 
         match_circ_with_simp_fns!(
-            (circuit, self.simp_fns());
+            (circuit, &self.1);
             (node, fns_sub) => run_simp(node, &fns_sub, on_name)
         )
     }
@@ -421,9 +443,8 @@ impl SimpFnSubset {
     }
 
     pub fn arg_fmt(f: impl Fn(&'static str) -> Option<String>, exclude_if_none: bool) -> String {
-        let fns = all_simp_fns();
         let tab = " ".repeat(4);
-        map_simp_fns_arr_name!(&fns; (sub_fns, circ_ty) => {
+        map_simp_fns_arr_name!(all_simp_fns(); (sub_fns, circ_ty) => {
             let sub_str = sub_fns
                 .iter()
                 .filter_map(|(n, _)| Some(format!("{}{}{},", tab, n, f(n)?)))
