@@ -50,17 +50,13 @@ pub fn generalfunction_special_case_simplification(circuit: &GeneralFunction) ->
                 let last_dim_maybe_symbolic =
                     circuit.nodes[0].info().shape[circuit.nodes[0].info().rank() - 1];
                 if !SymbolicSizeProduct::has_symbolic(last_dim_maybe_symbolic) {
-                    return Some(Array::nrc(
-                        scalar_to_tensor(
-                            last_dim_maybe_symbolic as f64,
-                            circuit.nodes[0].info().shape[..circuit.nodes[0].info().rank() - 1]
-                                .iter()
-                                .cloned()
-                                .collect(),
-                            circuit.info().device_dtype.unwrap_or_defaults(),
-                        )
-                        .unwrap(),
-                        None,
+                    return Some(Scalar::nrc(
+                        last_dim_maybe_symbolic as f64,
+                        circuit.nodes[0].info().shape[..circuit.nodes[0].info().rank() - 1]
+                            .iter()
+                            .cloned()
+                            .collect(),
+                        circuit.info().name,
                     ));
                 }
             }
