@@ -207,7 +207,7 @@ where
 pub fn circuit_short_print(circuit: &CircuitRc) -> String {
     format!(
         "{} {} {}",
-        match circuit.name() {
+        match circuit.info().name {
             None => "None".to_owned(),
             Some(s) => format!("'{}'", s),
         },
@@ -292,7 +292,7 @@ pub fn repr_matcher_debug(
                 result,
                 "{}'{}'...\n",
                 last_child_arrows(&last_child_stack, true, print_options.arrows),
-                circuit.name_cloned().unwrap_or("".to_owned())
+                circuit.info().name.unwrap_or("".into())
             )
             .unwrap();
             return Ok(());
@@ -381,12 +381,13 @@ pub fn append_matchers_to_names(
             .collect::<Result<Vec<CircuitRc>>>()?;
         let result = circuit.map_children_unwrap_idxs(|i| new_children[i].clone());
         if discard_old_name {
-            return Ok(result.rename(Some(format!("{}", &matcher.0))));
+            return Ok(result.rename(Some(format!("{}", &matcher.0).into())));
         }
         Ok(result.rename(
             circuit
-                .name()
-                .map(|n| format!("{} Matcher {}", n, &matcher.0)),
+                .info()
+                .name
+                .map(|n| format!("{} Matcher {}", n, &matcher.0).into()),
         ))
     }
     recurse(circuit, Some(matcher), discard_old_name)
