@@ -1,9 +1,11 @@
 use std::{
+    iter::zip,
     num::NonZeroUsize,
     ops::{Deref, DerefMut},
 };
 
 use pyo3::{FromPyObject, IntoPy, PyAny, PyObject, PyResult, Python};
+use rustc_hash::FxHashSet as HashSet;
 
 /// stores 7 elements inline
 // This is 8 bytes long and stores either 7 u8s or an owning pointer (like Box) to a Vec<u8>
@@ -30,14 +32,17 @@ impl TinyVecU8 {
     unsafe fn pointer(&self) -> &Box<Vec<u8>> {
         std::mem::transmute::<&NonZeroUsize, &Box<Vec<u8>>>(&self.0)
     }
+    #[inline]
     unsafe fn inline_len(&self) -> usize {
         (self.raw()[0] >> 1) as usize
     }
+    #[inline]
     unsafe fn set_inline_len(&mut self, value: usize) {
         self.raw_mut()[0] = (value as u8) << 1 | 0b1;
     }
 
     // safe api stuff
+    #[inline]
     pub fn is_inline(&self) -> bool {
         unsafe { self.raw()[0] & 0b1 != 0 }
     }
@@ -217,6 +222,124 @@ macro_rules! tu8v{
     };
 }
 
+#[derive(Default, Copy, Clone)]
+pub struct U8Set([u64; 4]); // 64 bits * 4 = 256 bits
+impl U8Set {
+    #[inline]
+    pub fn contains(&self, x: u8) -> bool {
+        self.0[(x >> 6) as usize] & (1 << (x & 0b00111111)) != 0
+    }
+    #[inline]
+    pub fn insert(&mut self, x: u8) {
+        self.0[(x >> 6) as usize] |= 1 << (x & 0b00111111);
+    }
+    pub fn len(&self) -> usize {
+        self.0.into_iter().map(|x| x.count_ones() as usize).sum() // portable_simd doesn't have count_ones
+    }
+    pub fn is_subset(&self, other: &U8Set) -> bool {
+        zip(self.0, other.0).all(|(sub, sup)| sub & !sup == 0) // could use simd
+    }
+    pub fn is_superset(&self, other: &U8Set) -> bool {
+        other.is_subset(&self)
+    }
+    pub fn union(&self, other: &U8Set) -> U8Set {
+        Self([
+            self.0[0] | other.0[0],
+            self.0[1] | other.0[1],
+            self.0[2] | other.0[2],
+            self.0[3] | other.0[3],
+        ])
+    }
+    pub fn intersection(&self, other: &U8Set) -> U8Set {
+        Self([
+            self.0[0] & other.0[0],
+            self.0[1] & other.0[1],
+            self.0[2] & other.0[2],
+            self.0[3] & other.0[3],
+        ])
+    }
+}
+
+impl FromIterator<u8> for U8Set {
+    fn from_iter<I: IntoIterator<Item = u8>>(iter: I) -> U8Set {
+        let mut c = U8Set::default();
+
+        for i in iter {
+            c.insert(i);
+        }
+
+        c
+    }
+}
+pub struct U8SetIterator {
+    u8set: U8Set,
+    chunk: usize,
+}
+impl Iterator for U8SetIterator {
+    type Item = u8;
+    fn next(&mut self) -> Option<Self::Item> {
+        loop {
+            match self.u8set.0[self.chunk].trailing_zeros() {
+                64 => {
+                    if self.chunk == 3 {
+                        return None;
+                    }
+                    self.chunk += 1;
+                }
+                num_zeros => {
+                    self.u8set.0[self.chunk] &= !(1 << num_zeros);
+                    return Some(num_zeros as u8 + 64 * self.chunk as u8);
+                }
+            }
+        }
+    }
+}
+
+impl IntoIterator for U8Set {
+    type Item = u8;
+    type IntoIter = U8SetIterator;
+    fn into_iter(self) -> Self::IntoIter {
+        U8SetIterator {
+            u8set: self,
+            chunk: 0,
+        }
+    }
+}
+
+impl std::fmt::Debug for U8Set {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        self.into_iter().collect::<HashSet<u8>>().fmt(f)
+    }
+}
+impl IntoPy<PyObject> for U8Set {
+    fn into_py(self, py: Python<'_>) -> PyObject {
+        self.into_iter().collect::<HashSet<u8>>().into_py(py)
+    }
+}
+impl<'source> FromPyObject<'source> for U8Set {
+    fn extract(args_obj: &'source PyAny) -> PyResult<Self> {
+        let vec: HashSet<u8> = args_obj.extract()?;
+
+        Ok(vec.into_iter().collect())
+    }
+}
+#[test]
+fn test_u8set() {
+    let mut u8s = U8Set::default();
+    dbg!(u8s.len());
+    u8s.insert(0);
+    u8s.insert(243);
+    u8s.insert(150);
+    u8s.insert(255);
+    assert!(u8s.into_iter().collect::<Vec<u8>>()[..] == vec![0, 150, 243, 255]);
+    assert!(u8s.len() == 4);
+    dbg!(u8s.into_iter().collect::<Vec<u8>>(), u8s.len());
+    assert!(u8s.contains(150) && !u8s.contains(1));
+    let mut u8s2: U8Set = vec![0, 1, 2, 2, 50].into_iter().collect();
+    assert!(u8s2.union(&u8s).into_iter().collect::<Vec<u8>>() == vec![0, 1, 2, 50, 150, 243, 255]);
+    dbg!(u8s2.intersection(&u8s).into_iter().collect::<Vec<u8>>() == vec![0]);
+}
+
 #[test]
 fn test_tiny_u8() {
     let mut col = TinyVecU8::new();
