@@ -1,7 +1,8 @@
-use std::{collections::HashMap, fmt::Debug, hash::Hash, ptr};
+use std::{fmt::Debug, hash::Hash, ptr};
 
 use anyhow::Result;
 use pyo3::{prelude::*, types::IntoPyDict};
+use rustc_hash::FxHashMap as HashMap;
 
 use crate::{py_types::Tensor, tensor_db::get_tensor_prefix, tensor_util::TorchDeviceDtype};
 // implement our own weighted LRU cache bc didnt see a weighted library
