@@ -56,6 +56,13 @@ impl OpSize {
         assert!(self.is_some());
         self.0 as usize
     }
+    pub fn unwrap_or(self, x: usize) -> usize {
+        if self.is_none() {
+            x
+        } else {
+            self.unwrap()
+        }
+    }
 }
 
 impl Debug for OpSize {
@@ -254,7 +261,7 @@ impl FromStr for RearrangeSpec {
 /// must be specified as int_sizes[i] = Some(repeats).
 /// If a variable is on the left, it must appear on the right (we don't allow reductions). This means
 /// this doesn't support squeezes normally: '1 a -> a' is not allowed, but unsqueezes are allowed.
-/// (You can however get a squeeze by doing '1 a -> (1 a)', although you should use Index instead.)
+/// (You can however get a squeeze by doing '() a -> a' or `b:1 a -> (b:1 a)`, although you should use Index instead.)
 #[pymethods]
 impl RearrangeSpec {
     #[new]
@@ -930,6 +937,10 @@ impl RearrangeSpec {
     fn conform_to_input_shape_py(&self, shape: Shape) -> Result<RearrangeSpec> {
         self.conform_to_input_shape(&shape)
     }
+    #[pyo3(name = "conform_to_output_shape")]
+    fn conform_to_output_shape_py(&self, shape: Shape) -> Result<RearrangeSpec> {
+        self.conform_to_output_shape(&shape)
+    }
 
     pub fn next_axis(&self) -> u8 {
         // input_ints are included in output_ints, so we don't check them
@@ -993,7 +1004,7 @@ impl RearrangeSpec {
         error_on_ambiguous: bool,
     ) -> Result<ExpandToSpecOrShape> {
         if shape.len() != self.input_ints.len() {
-            bail!(RearrangeSpecError::InputNotConformable {
+            bail!(RearrangeSpecError::NotConformable {
                 spec: self.clone(),
                 shape: shape.clone(),
             });
@@ -1072,54 +1083,86 @@ impl RearrangeSpec {
 
     pub fn conform_to_input_shape(&self, shape: &Shape) -> Result<RearrangeSpec> {
         if shape.len() != self.input_ints.len() {
-            bail!(RearrangeSpecError::InputNotConformable {
+            bail!(RearrangeSpecError::NotConformable {
                 spec: self.clone(),
                 shape: shape.clone(),
             });
         }
         let mut int_sizes = self.int_sizes.clone();
-        for (ints, l) in zip(&self.input_ints, shape) {
+        for (ints, &l) in zip(&self.input_ints, shape) {
             let none_indices: Vec<usize> = ints
                 .iter()
                 .enumerate()
                 .filter(|(_i, x)| self.int_sizes[**x as usize].is_none())
                 .map(|(i, _x)| i)
                 .collect();
-            let non_wildcard_size: u64 = ints
+            let non_wildcard_size: usize = ints
                 .iter()
-                .map(|x| {
-                    let size_here = self.int_sizes[*x as usize];
-                    if size_here.is_none() {
-                        1
-                    } else {
-                        size_here.0
-                    }
-                })
+                .map(|x| self.int_sizes[*x as usize].unwrap_or(1))
                 .product();
-            // dbg!(&int_sizes,&non_wildcard_size,&ints,&l);
-            if non_wildcard_size != *l as u64 {
+            let has_wildcard = none_indices.len() > 0;
+            if (!has_wildcard && l != non_wildcard_size)
+                || (has_wildcard && l % non_wildcard_size != 0)
+            {
+                bail!(RearrangeSpecError::NotConformable {
+                    spec: self.clone(),
+                    shape: shape.clone(),
+                });
+            }
+
+            if has_wildcard {
                 if none_indices.len() > 1 {
                     bail!(RearrangeSpecError::TooManyWildcardSizes { spec: self.clone() });
-                } else if *l as u64 % non_wildcard_size != 0 {
-                    bail!(RearrangeSpecError::InputNotConformable {
-                        spec: self.clone(),
-                        shape: shape.clone(),
-                    });
                 } else if none_indices.len() == 1 {
                     int_sizes[ints[none_indices[0]] as usize] =
-                        OpSize(*l as u64 / non_wildcard_size);
-                } else {
-                    bail!(RearrangeSpecError::InputNotConformable {
-                        spec: self.clone(),
-                        shape: shape.clone(),
-                    });
+                        OpSize((l / non_wildcard_size) as u64);
                 }
-            } else if ints.len() == 1 {
-                int_sizes[ints[0] as usize] = OpSize(*l as u64);
             }
         }
         Ok(Self::new(self.input_ints.clone(), self.output_ints.clone(), int_sizes).unwrap())
     }
+
+    pub fn conform_to_output_shape(&self, shape: &Shape) -> Result<RearrangeSpec> {
+        if shape.len() != self.output_ints.len() {
+            bail!(RearrangeSpecError::NotConformable {
+                spec: self.clone(),
+                shape: shape.clone(),
+            });
+        }
+        let mut int_sizes = self.int_sizes.clone();
+        for (ints, &l) in zip(&self.output_ints, shape) {
+            let none_indices: Vec<usize> = ints
+                .iter()
+                .enumerate()
+                .filter(|(_i, x)| self.int_sizes[**x as usize].is_none())
+                .map(|(i, _x)| i)
+                .collect();
+            let non_wildcard_size: usize = ints
+                .iter()
+                .map(|x| self.int_sizes[*x as usize].unwrap_or(1))
+                .product();
+            let has_wildcard = none_indices.len() > 0;
+            if (!has_wildcard && l != non_wildcard_size)
+                || (has_wildcard && l % non_wildcard_size != 0)
+            {
+                bail!(RearrangeSpecError::NotConformable {
+                    spec: self.clone(),
+                    shape: shape.clone(),
+                });
+            }
+
+            if has_wildcard {
+                if none_indices.len() > 1 {
+                    bail!(RearrangeSpecError::TooManyWildcardSizes { spec: self.clone() });
+                } else if none_indices.len() == 1 {
+                    int_sizes[ints[none_indices[0]] as usize] =
+                        OpSize((l / non_wildcard_size) as u64);
+                }
+            }
+        }
+        Ok(Self::new(self.input_ints.clone(), self.output_ints.clone(), int_sizes).unwrap())
+    }
+
     pub fn unflatten_axis_usize(ndim: usize, axis: usize, shape: Shape) -> Result<Self> {
         Self::unflatten_axis(ndim, axis as i64, shape)
     }
@@ -1601,7 +1644,7 @@ fn fancy_size_format<'a, T: IntoIterator<Item = &'a usize>>(x: T) -> String {
 #[derive(Error, Debug, Clone)]
 pub enum RearrangeSpecError {
     #[error("RearrangeSpec not conformable, {spec:?} {shape:?} ({e_name})")]
-    InputNotConformable { spec: RearrangeSpec, shape: Shape },
+    NotConformable { spec: RearrangeSpec, shape: Shape },
 
     #[error("RearrangeSpec has wildcard sizes, {spec:?} ({e_name})")]
     HasWildcardSizes { spec: RearrangeSpec },
